from typing import List, Set, Tuple

from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.db.converter import Converter
from neuralogic.db.pg.helpers import helpers


FUNCTION_TEMPLATE = """
CREATE OR REPLACE FUNCTION {name}({params}) RETURNS {return_type} AS $$
{body}
$$ LANGUAGE {language} {volatility};
""".strip()


FUNCTION_MAP = {
    "mul": "neuralogic_std.mul",
    "sum": "neuralogic_std.sum",
    "tanh": "neuralogic_std.tanh",
    "sigmoid": "neuralogic_std.sigmoid",
    "relu": "neuralogic_std.relu",
    "identity": "",
    "avg": "AVG",
    "max": "MAX",
    "min": "MIN",
}


class PostgresConverter(Converter):
    @staticmethod
    def get_function(
        name: str,
        params: List[str],
        return_type: List[str],
        body: str,
        language: str = "SQL",
        volatility: str = "STABLE",
    ) -> str:
        return FUNCTION_TEMPLATE.format(
            name=name,
            params=",".join(params),
            return_type=f"Table({','.join(return_type)})",
            body=body,
            language=language,
            volatility=volatility,
        )

    @staticmethod
    def get_empty_function(
        name: str,
        params: List[str],
        return_type: List[str],
        language: str = "SQL",
        volatility: str = "STABLE",
    ) -> str:
        if len(return_type) == 0:
            raise NotImplementedError

        select = ["1"]
        for _ in return_type[1:]:
            select.append("'1'")

        return PostgresConverter.get_function(
            name, params, return_type, f"SELECT {','.join(select)}", language, volatility
        )

    def get_helpers(self, functions: Set[str]) -> str:
        used_functions = {fun for fun in functions}
        used_functions.add("mul")
        used_functions.add("sum")

        function_sources = ["CREATE SCHEMA IF NOT EXISTS neuralogic_std;" "CREATE SCHEMA IF NOT EXISTS neuralogic;"]

        for fun in used_functions:
            helper_fun = helpers[fun]

            if helper_fun is not None:
                function_sources.append(helper_fun.strip())
        return "\n".join(function_sources)

    def get_fact_sql_function(self, relation: BaseRelation, index: int, weight_indices: List[int], weights) -> str:
        """Generate a SQL function for a ground fact"""
        value = 1 if weight_indices[0] is None else weights[weight_indices[0]]

        parameters = [f"p{i} TEXT" for i in range(len(relation.terms))]
        return_type = ["value NUMERIC", *(f"t{i} TEXT" for i in range(len(relation.terms)))]
        select = [f"{value} as value", *(f"'{term}' as t{i}" for i, term in enumerate(relation.terms))]

        name = f"neuralogic._{relation.predicate.name}_{relation.predicate.arity}_{index}"
        body = f"SELECT {','.join(select)}"
        condition = " AND ".join(f"(p{i} = '{term}' OR p{i} IS NULL)" for i, term in enumerate(relation.terms))

        if condition:
            body = f"{body} WHERE {condition}"

        return self.get_function(name, parameters, return_type, body, volatility="IMMUTABLE")

    def get_relation_interface_sql_function(self, relation: str, arity: int) -> Tuple[str, str]:
        """Return the SQL function that should by used by the end users"""
        function_parameters = [f"p{i}" for i in range(arity)]
        params = [f"{name} TEXT" for name in function_parameters]

        name = f"neuralogic.{relation}"
        return_type = ["value NUMERIC", *(f"t{i} TEXT" for i in range(arity))]
        function_name = f"neuralogic._{relation}_{arity}({','.join('NULL' for _ in range(arity))})"
        tmp_table_name = f"__neuralogic___tmp_{relation}_{arity}"
        conditions = " AND ".join(f"out.t{i} LIKE COALESCE(p{i}, '%')" for i in range(arity))

        if conditions:
            return_select = f"RETURN QUERY SELECT * FROM {tmp_table_name} as out WHERE {conditions}"
        else:
            return_select = f"RETURN QUERY SELECT * FROM {tmp_table_name} as out"

        body = (
            f"BEGIN CREATE TEMPORARY TABLE IF NOT EXISTS {tmp_table_name} ON COMMIT DROP AS "
            f"SELECT * FROM {function_name}; {return_select}; RETURN; END;"
        )

        return (
            self.get_empty_function(name, params, return_type, "SQL", "VOLATILE"),
            self.get_function(name, params, return_type, body, "plpgsql", "VOLATILE"),
        )

    def get_rule_aggregation_function(
        self, name: str, arity: int, number_of_rules: int, activation: str, aggregation: str, is_fact: bool = False
    ) -> str:
        """
        Generete SQL function which aggregates rule functions (something like the aggregation neuron)
        """
        function_parameters = [f"p{i}" for i in range(arity)]

        inner_select = []

        for index in range(number_of_rules):
            function_name = f"neuralogic._{name}_{arity}_{index}({','.join(function_parameters)}) as s{index}"
            selects = [f"s{index}.value as value"]
            selects.extend(f"s{index}.t{i}" for i in range(arity))

            selects = ", ".join(selects)
            inner_select.append(f"SELECT {selects} FROM {function_name}")

        if is_fact:
            select = ["SUM(out.value) as value"]
        else:
            select = [f"{FUNCTION_MAP[activation]}(SUM(out.value)) as value"]

        select.extend(f"out.t{i}" for i in range(arity))
        select = ", ".join(select)

        group_by_clause = f" GROUP BY {', '.join('out.t' + str(v) for v in range(arity))}"
        select_from = " UNION ALL ".join(inner_select)

        return_type = ["value NUMERIC", *(f"t{i} TEXT" for i in range(arity))]
        name = f"neuralogic._{name}_{arity}"
        body = f"SELECT {select} FROM ({select_from}) AS out{'' if arity == 0 else group_by_clause}"

        return self.get_function(name, [f"{name} TEXT" for name in function_parameters], return_type, body)

    def get_rule_sql_function(
        self, rule: Rule, index: int, activation: str, aggregation: str, weight_indices: List[int], weights
    ) -> str:
        """Return the SQL function of one rule"""
        function_parameters = [f"p{i}" for i in range(rule.head.predicate.arity)]

        if weight_indices[0] is None:
            select = [f"{FUNCTION_MAP[aggregation]}({FUNCTION_MAP[activation]}(out.value)) as value"]
        else:
            select = [
                f"{FUNCTION_MAP['mul']}({weights[weight_indices[0]]}, "
                f"{FUNCTION_MAP[aggregation]}({FUNCTION_MAP[activation]}(out.value))) as value"
            ]

        vars_mapping = {}
        where = []
        vars_body_mapping = {}
        join_vars_mapping = {}
        inner_select = set()
        inner_selected_terms = set()
        inner_value_select = []
        from_clause: List[str] = []

        for term_idx, term in enumerate(rule.head.terms):
            if Converter._is_var(term):
                term_name = f"t{term_idx}"
                vars_mapping[str(term)] = term_name
                vars_body_mapping[str(term)] = f"p{term_idx}"

                select.append(f"out.{term_name}")
            else:
                select.append(f"'{term}' as t{term_idx}")

        for t_index, (relation, weight_id) in enumerate(zip(rule.body, weight_indices[1:])):
            join_on = []

            relation_mapping = self.table_mappings.get(str(relation.predicate).replace("*", "_"), None)

            if not relation.predicate.hidden:
                if relation_mapping is None:
                    value = "value"
                else:
                    value_column = relation_mapping.value_column
                    value = None if value_column is None else f"{value_column}::NUMERIC"
                selected_value = "1" if value is None else f"s{t_index}.{value}"

                if relation.function is not None:
                    function = str(relation.function).lower()
                    selected_value = f"{FUNCTION_MAP[function]}({selected_value})"

                    self._used_functions.add(function)

                if weight_id is None:
                    inner_value_select.append(selected_value)
                else:
                    inner_value_select.append(f"{FUNCTION_MAP['mul']}({weights[weight_id]}, {selected_value})")

                if len(inner_value_select) == 2:
                    inner_value_select = [f"{FUNCTION_MAP['sum']}({', '.join(inner_value_select)})"]

            from_function_parameters = []
            for term_idx, term in enumerate(relation.terms):
                if relation_mapping is None:
                    field = f"t{term_idx}"
                else:
                    field = f"{relation_mapping.term_columns[term_idx]}::TEXT"

                if not self._is_var(term):
                    if relation_mapping is None:
                        from_function_parameters.append(f"'{term}'")
                    else:
                        where.append(f"s{t_index}.{field} = '{str(term)}'")
                    continue

                from_function_parameters.append("NULL")

                if str(term) in vars_mapping and str(term) not in inner_selected_terms:
                    inner_select.add(f"s{t_index}.{field} AS {vars_mapping[str(term)]}")
                    inner_selected_terms.add(str(term))

                if str(term) in vars_body_mapping:
                    where.append(f"s{t_index}.{field} LIKE COALESCE({vars_body_mapping[str(term)]}, '%')")

                    if str(term) in join_vars_mapping:
                        join_on.append(f"{vars_body_mapping[str(term)]} = s{t_index}.{field}")
                    else:
                        vars_body_mapping[str(term)] = f"s{t_index}.{field}"
                        join_vars_mapping[str(term)] = f"s{t_index}.{field}"
                    continue

                join_vars_mapping[str(term)] = f"s{t_index}.{field}"
                vars_body_mapping[str(term)] = f"s{t_index}.{field}"

            if relation_mapping is None:
                params = ",".join(from_function_parameters)
                function_name = f"neuralogic.{relation.predicate.name}({params})"
            else:
                function_name = relation_mapping.table_name

            if not from_clause:
                from_clause.append(f"{function_name} AS s{t_index}")
            else:
                from_clause.append(
                    f"{function_name} AS s{t_index} ON {'1 = 1' if not join_on else ' AND '.join(join_on)}"
                )

        from_clause = f"{' INNER JOIN '.join(from_clause)}"
        where_clause = f" WHERE {' AND '.join(where)}"
        group_by_clause = f" GROUP BY {', '.join('out.' + v for v in vars_mapping.values())}"

        if len(inner_value_select) == 1:
            inner_select.add(f"{inner_value_select[0]} as value")

        from_clause = f"SELECT {', '.join(inner_select)} FROM {from_clause}{'' if not where else where_clause}"

        return_type = ["value NUMERIC", *(f"t{i} TEXT" for i in range(len(rule.head.terms)))]

        name = f"neuralogic._{rule.head.predicate.name}_{rule.head.predicate.arity}_{index}"
        body = f"SELECT {', '.join(select)} FROM ({from_clause}) AS out{'' if not vars_mapping else group_by_clause}"

        return self.get_function(name, [f"{name} TEXT" for name in function_parameters], return_type, body)



import jpype

import neuralogic
from neuralogic import is_initialized, initialize
from neuralogic.core.constructs.function import Transformation, Combination
from neuralogic.core.enums import Grounder
from neuralogic.nn.init import Initializer
from neuralogic.nn.loss import MSE, SoftEntropy, CrossEntropy, ErrorFunction
from neuralogic.optim import Optimizer


class SettingsProxy:
    def __init__(
        self,
        *,
        optimizer: Optimizer,
        learning_rate: float,
        epochs: int,
        error_function: ErrorFunction,
        initializer: Initializer,
        rule_transformation: Transformation,
        rule_combination: Combination,
        relation_transformation: Transformation,
        relation_combination: Combination,
        iso_value_compression: bool,
        chain_pruning: bool,
        prune_only_identities: bool,
        grounder: Grounder,
    ):
        if not is_initialized():
            initialize()

        self.settings_class = jpype.JClass("cz.cvut.fel.ida.setup.Settings")
        self.settings = self.settings_class()

        self._optimizer = optimizer

        params = locals().copy()
        params.pop("self")

        for key, value in params.items():
            self.__setattr__(key, value)

        self.settings.debugExporting = False
        self.settings.exportBlocks = []

        self.settings.infer()
        self._setup_random_generator()

    def _setup_random_generator(self):
        if neuralogic._rnd_generator is None:
            neuralogic._rnd_generator = self.settings.random
            self.settings.random.setSeed(neuralogic._seed)
        self.settings.random = neuralogic._rnd_generator

    @property
    def iso_value_compression(self) -> bool:
        return self.settings.isoValueCompression

    @iso_value_compression.setter
    def iso_value_compression(self, iso_value_compression: bool):
        self.settings.isoValueCompression = iso_value_compression

    @property
    def chain_pruning(self) -> bool:
        return self.settings.chainPruning

    @chain_pruning.setter
    def chain_pruning(self, chain_pruning: bool):
        self.settings.chainPruning = chain_pruning

    @property
    def prune_only_identities(self) -> bool:
        return self.settings.pruneOnlyIdentities

    @prune_only_identities.setter
    def prune_only_identities(self, prune_only_identities: bool):
        self.settings.pruneOnlyIdentities = prune_only_identities

    @property
    def grounder(self):
        return self.settings.grounding

    @grounder.setter
    def grounder(self, grounder: Grounder):
        if grounder == Grounder.BUP:
            self.settings.grounding = self.settings.GroundingAlgo.BUP
        elif grounder == Grounder.GRINGO:
            self.settings.grounding = self.settings.GroundingAlgo.GRINGO
        else:
            raise ValueError(f"Invalid grounder {grounder}")

    @property
    def optimizer(self) -> Optimizer:
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        if optimizer.name() == "SGD":
            java_optimizer = self.settings_class.OptimizerSet.SGD
        elif optimizer.name() == "Adam":
            java_optimizer = self.settings_class.OptimizerSet.ADAM
        else:
            raise NotImplementedError

        self._optimizer = optimizer
        self.settings.setOptimizer(java_optimizer)
        self.settings.initLearningRate = optimizer.lr

    @property
    def initializer_const(self):
        return self.settings.constantInitValue

    @initializer_const.setter
    def initializer_const(self, value: float):
        self.settings.constantInitValue = value

    @property
    def initializer_uniform_scale(self):
        return self.settings.randomInitScale

    @initializer_uniform_scale.setter
    def initializer_uniform_scale(self, value: float):
        self.settings.randomInitScale = value

    @property
    def error_function(self):
        return self.settings.errorFunction

    @error_function.setter
    def error_function(self, error_function: ErrorFunction):
        self.settings.inferOutputFcns = True

        if isinstance(error_function, MSE):
            self.settings.squishLastLayer = False
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.REGRESSION
            java_error_function = self.settings_class.ErrorFcn.SQUARED_DIFF
        elif isinstance(error_function, SoftEntropy):
            self.settings.squishLastLayer = True
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.CLASSIFICATION
            java_error_function = self.settings_class.ErrorFcn.SOFTENTROPY
        elif isinstance(error_function, CrossEntropy):
            self.settings.trainOnlineResultsType = self.settings_class.ResultsType.CLASSIFICATION

            if error_function.with_logits:
                self.settings.squishLastLayer = True
                java_error_function = self.settings_class.ErrorFcn.SOFTENTROPY
            else:
                self.settings.inferOutputFcns = False
                self.settings.squishLastLayer = False
                java_error_function = self.settings_class.ErrorFcn.CROSSENTROPY
        else:
            raise NotImplementedError

        self.settings.errorFunction = java_error_function

    @property
    def epochs(self) -> int:
        return self.settings.maxCumEpochCount

    @epochs.setter
    def epochs(self, epochs: int):
        self.settings.maxCumEpochCount = epochs

    @property
    def initializer(self):
        initializer = self.settings.initializer

        if str(initializer) != "SIMPLE":
            return initializer
        return self.settings.initDistribution

    @initializer.setter
    def initializer(self, initializer: Initializer):
        if not isinstance(initializer, Initializer):
            raise TypeError()

        settings = initializer.get_settings()
        init_name = settings.pop("initializer")

        if initializer.is_simple():
            self.settings.initializer = self.settings_class.InitSet.SIMPLE
            self.settings.initDistribution = getattr(self.settings_class.InitDistribution, init_name)
        else:
            self.settings.initializer = getattr(self.settings_class.InitSet, init_name)

        for key, value in settings.items():
            self.__setattr__(key, value)

    @property
    def relation_transformation(self) -> Transformation:
        return Transformation(str(self.settings.atomNeuronTransformation))

    @relation_transformation.setter
    def relation_transformation(self, value: Transformation):
        self.settings.atomNeuronTransformation = self.get_transformation_function(value)

    @property
    def relation_combination(self) -> Combination:
        return Combination(str(self.settings.atomNeuronCombination))

    @relation_combination.setter
    def relation_combination(self, value: Combination):
        self.settings.atomNeuronCombination = self.get_combination_function(value)

    @property
    def rule_transformation(self) -> Transformation:
        return Transformation(str(self.settings.ruleNeuronTransformation))

    @rule_transformation.setter
    def rule_transformation(self, value: Transformation):
        self.settings.ruleNeuronTransformation = self.get_transformation_function(value)

    @property
    def rule_combination(self) -> Combination:
        return Combination(str(self.settings.ruleNeuronCombination))

    @rule_combination.setter
    def rule_combination(self, value: Combination):
        self.settings.ruleNeuronCombination = self.get_combination_function(value)

    @property
    def debug_exporting(self) -> bool:
        return self.settings.debugExporting

    @debug_exporting.setter
    def debug_exporting(self, debug_export: bool):
        self.settings.debugExporting = debug_export

    @property
    def default_fact_value(self) -> float:
        return self.settings.defaultFactValue

    @default_fact_value.setter
    def default_fact_value(self, value: float):
        self.settings.defaultFactValue = value

    def get_combination_function(self, combination: Combination):
        combination_name = str(combination)
        return self.settings_class.parseCombination(combination_name)

    def get_transformation_function(self, transformation: Transformation):
        transformation_name = str(transformation)
        return self.settings_class.parseTransformation(transformation_name)

    def to_json(self) -> str:
        return self.settings.exportToJson()



def get_neuralogic_layer():
    from neuralogic.nn.java import NeuraLogic  # type: ignore

    return NeuraLogic


def get_evaluator(
    template,
    settings=None,
):
    from neuralogic.nn.evaluator.java import JavaEvaluator
    from neuralogic.core.settings import Settings

    if settings is None:
        settings = Settings()
    return JavaEvaluator(template, settings)



class Function:
    __slots__ = ("name",)

    def __init__(self, name: str):
        self.name: str = name.lower()

    def __str__(self):
        return self.name

    def pretty_str(self) -> str:
        return str(self).capitalize()

    def __call__(self, *args):
        if len(args) == 0 or args[0] is None:
            return self
        raise NotImplementedError


class Transformation(Function):
    # Element wise
    SIGMOID: "Transformation"
    TANH: "Transformation"
    SIGNUM: "Transformation"
    RELU: "Transformation"
    LEAKY_RELU: "Transformation"
    LUKASIEWICZ: "Transformation"
    EXP: "Transformation"
    SQRT: "Transformation"
    INVERSE: "Transformation"
    REVERSE: "Transformation"
    LOG: "Transformation"

    # Transformation
    IDENTITY: "Transformation"
    TRANSP: "Transformation"
    SOFTMAX: "Transformation"
    SPARSEMAX: "Transformation"
    NORM: "Transformation"

    def __call__(self, *args):
        from neuralogic.core.constructs import relation

        if len(args) == 0 or args[0] is None:
            return self

        arg = args[0]
        if isinstance(arg, relation.BaseRelation):
            return arg.attach_activation_function(self)
        raise NotImplementedError


_special_namings = {"LEAKY_RELU": "LEAKYRELU", "TRANSP": "TRANSPOSE"}

for function_name in Transformation.__annotations__:
    setattr(Transformation, function_name, Transformation(_special_namings.get(function_name, function_name)))


class Combination(Function):
    # Aggregation
    AVG: "Combination"
    MAX: "Combination"
    MIN: "Combination"
    SUM: "Combination"
    COUNT: "Combination"

    # Combination
    PRODUCT: "Combination"
    ELPRODUCT: "Combination"
    SOFTMAX: "Combination"
    SPARSEMAX: "Combination"
    CROSSSUM: "Combination"
    CONCAT: "Combination"
    COSSIM: "Combination"


for function_name in Combination.__annotations__:
    setattr(Combination, function_name, Combination(function_name))


class Aggregation(Function):
    AVG: "Aggregation"
    MAX: "Aggregation"
    MIN: "Aggregation"
    SUM: "Aggregation"
    COUNT: "Aggregation"


for function_name in Aggregation.__annotations__:
    setattr(Aggregation, function_name, Aggregation(function_name))



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class RNNCell(Module):
    r"""

    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_input_name : str
        Predicate name to get hidden state from.
    activation : Transformation
        Activation function.
        Default: ``Transformation.TANH``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_name: str,
        input_name: str,
        hidden_input_name: str,
        activation: Transformation = Transformation.TANH,
        arity: int = 1,
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_input_name = hidden_input_name

        self.activation = activation
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        output = R.get(self.output_name)
        input_terms = [*terms, V.T]

        rnn_rule = output([*terms, V.T]) <= (
            R.get(self.input_name)(input_terms)[self.hidden_size, self.input_size],
            R.get(self.hidden_input_name)([*terms, V.Z])[self.hidden_size, self.hidden_size],
            R.special.next(V.Z, V.T),
        )

        return [
            rnn_rule | Metadata(transformation=self.activation),
            output / (self.arity + 1) | [Transformation.IDENTITY],
        ]


class RNN(Module):
    r"""
    One-layer Recurrent Neural Network (RNN) module which is computed as:

    .. math::

        h_t = act(\mathbf{W}_{ih} \mathbf{x}_t + \mathbf{W}_{hh} \mathbf{h}_{t-1})

    where :math:`t \in (1, sequence\_length + 1)` is a time step.
    In the template, the :math:`t` is referred to as :code:`V.T`, and :math:`t - 1` is referred to as :code:`V.Z`.
    This module expresses the first equation as:

    .. code:: logtalk

        (R.<output_name>(<...terms>, V.T) <= (
            R.<input_name>(<...terms>, V.T)[<hidden_size>, <input_size>],
            R.<hidden_input_name>(<...terms>, V.Z)[<hidden_size>, <hidden_size>],
            R.special.next(V.Z, V.T),
        )) | [<activation>]

        R.<output_name> / <arity> + 1 | [Transformation.IDENTITY]

    Additionally, we define a rule for the "stop condition", that is:

    .. code:: logtalk

        (R.<output_name>(<...terms>, 0) <= R.<hidden_0_name>(<...terms>)) | [Transformation.IDENTITY]


    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_0_name : str
        Predicate name to get initial hidden state from.
    activation : Transformation
        Activation function.
        Default: ``Transformation.TANH``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_name: str,
        input_name: str,
        hidden_0_name: str,
        activation: Transformation = Transformation.TANH,
        arity: int = 1,
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_0_name = hidden_0_name

        self.activation = activation
        self.arity = arity

    def __call__(self):
        recursive_cell = RNNCell(
            self.input_size,
            self.hidden_size,
            self.output_name,
            self.input_name,
            self.output_name,
            self.activation,
            self.arity,
        )

        terms = [f"X{i}" for i in range(self.arity)]

        return [
            (R.get(self.output_name)([*terms, 0]) <= R.get(self.hidden_0_name)(terms)) | [Transformation.IDENTITY],
            *recursive_cell(),
        ]



import json
from typing import Callable, Any, List, Union

import torch
from torch import nn
from torch.autograd import Function

from neuralogic.core import Template, Settings
from neuralogic.core.constructs.java_objects import ValueFactory
from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.dataset import Dataset


class _NeuraLogicFunction(Function):
    @staticmethod
    def forward(ctx, mapping, value_factory, sample, model, number_format, dtype, *inputs):
        ctx.model = model
        ctx.sample = sample
        ctx.number_format = number_format
        ctx.dtype = dtype
        ctx.mapping = mapping

        for fact in mapping:
            sample.set_fact_value(fact, value_factory.get_value(fact.weight)[1])

        return torch.tensor(model(sample, train=False), dtype=dtype, requires_grad=True)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        model = ctx.model
        sample = ctx.sample
        number_format = ctx.number_format
        dtype = ctx.dtype

        backproper, weight_updater = model.backprop(sample, -grad_outputs[0].detach().numpy())
        state_index = backproper.stateIndex

        gradients = tuple(
            -torch.tensor(
                json.loads(
                    str(sample.get_fact(fact).getComputationView(state_index).getGradient().toString(number_format))
                ),
                dtype=dtype,
            ).reshape(fact.weight.shape)
            for fact in ctx.mapping
        )

        trainer = model.strategy.getTrainer()
        trainer.updateWeights(model.strategy.getCurrentModel(), weight_updater)
        trainer.invalidateSample(trainer.getInvalidation(), sample.java_sample)

        return (None, None, None, None, None, None, *gradients)


class NeuraLogic(nn.Module):
    def __init__(
        self,
        template: Template,
        input_facts: List[Union[BaseRelation, Rule]],
        output_relation: BaseRelation,
        to_logic: Callable,
        settings: Settings = None,
        dtype=torch.float32,
    ):
        super(NeuraLogic, self).__init__()

        if settings is None:
            settings = Settings()
        settings.iso_value_compression = False
        settings.chain_pruning = False

        self.dtype = dtype
        self.to_logic = to_logic

        self.model = template.build(settings)
        self.number_format = self.model.settings.settings_class.superDetailedNumberFormat

        dataset = Dataset([input_facts], [output_relation])
        self.sample = self.model.build_dataset(dataset, learnable_facts=True).samples[0]
        self.value_factory = ValueFactory()

        self.internal_weights = nn.Parameter(torch.empty((0,)))
        self.neuralogic_function = _NeuraLogicFunction

    def forward(self, *inputs, **kwargs):
        mapping = self.to_logic(*inputs, **kwargs)

        return _NeuraLogicFunction.apply(
            mapping,
            self.value_factory,
            self.sample,
            self.model,
            self.number_format,
            self.dtype,
            *(fact.weight for fact in mapping),
        )



MUL = """
CREATE OR REPLACE FUNCTION neuralogic_std.mul(in_a NUMERIC, in_b NUMERIC) RETURNS NUMERIC AS
$$
SELECT (in_a * in_b)::NUMERIC
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.mul(in_a NUMERIC, in_b NUMERIC[]) RETURNS NUMERIC[] AS
$$
SELECT array_agg(in_a * un_b) FROM unnest(in_b) as un_b
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.mul(in_b NUMERIC[], in_a NUMERIC) RETURNS NUMERIC[] AS
$$
SELECT neuralogic_std.mul(in_a, in_b)
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.mul(in_a NUMERIC[], in_b NUMERIC[]) RETURNS NUMERIC AS
$$
SELECT SUM(un_a * un_b)::NUMERIC FROM unnest(in_a, in_b) as t(un_a, un_b)
$$
LANGUAGE SQL IMMUTABLE;
"""

SUM = """

CREATE OR REPLACE FUNCTION neuralogic_std.sum(in_a NUMERIC, in_b NUMERIC) RETURNS NUMERIC AS
$$
SELECT (in_a + in_b)::NUMERIC
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.sum(in_a NUMERIC, in_b NUMERIC[]) RETURNS NUMERIC[] AS
$$
SELECT array_agg(in_a + un_b) FROM unnest(in_b) as un_b
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.sum(in_b NUMERIC[], in_a NUMERIC) RETURNS NUMERIC[] AS
$$
SELECT neuralogic_std.sum(in_a, in_b)
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.sum(in_a NUMERIC[], in_b NUMERIC[]) RETURNS NUMERIC AS
$$
SELECT SUM(un_a + un_b)::NUMERIC FROM unnest(in_a, in_b) as t(un_a, un_b)
$$
LANGUAGE SQL IMMUTABLE;
"""

TANH = """
CREATE OR REPLACE FUNCTION neuralogic_std.tanh(in_a NUMERIC[]) RETURNS NUMERIC[] AS
$$
SELECT array_agg(tanh(un_a)) FROM unnest(in_a) as un_a
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.tanh(in_a NUMERIC) RETURNS NUMERIC AS
$$
SELECT tanh(in_a)
$$
LANGUAGE SQL IMMUTABLE;
"""

SIGMOID = """
CREATE OR REPLACE FUNCTION neuralogic_std.sigmoid(in_a NUMERIC) RETURNS NUMERIC AS
$$
SELECT 1 / (exp(-in_a) + 1)
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.sigmoid(in_a NUMERIC[]) RETURNS NUMERIC[] AS
$$
SELECT array_agg(neuralogic_std.sigmoid(un_a)) FROM unnest(in_a) as un_a
$$
LANGUAGE SQL IMMUTABLE;
"""

RELU = """
CREATE OR REPLACE FUNCTION neuralogic_std.relu(in_a NUMERIC) RETURNS NUMERIC AS
$$
SELECT CASE WHEN in_a > 0 THEN in_a ELSE 0 END
$$
LANGUAGE SQL IMMUTABLE;

CREATE OR REPLACE FUNCTION neuralogic_std.relu(in_a NUMERIC[]) RETURNS NUMERIC[] AS
$$
SELECT array_agg(neuralogic_std.relu(un_a)) FROM unnest(in_a) as un_a
$$
LANGUAGE SQL IMMUTABLE;
"""


helpers = {
    "mul": MUL.replace("\n\n", "\n"),
    "sum": SUM.replace("\n\n", "\n"),
    "tanh": TANH.replace("\n\n", "\n"),
    "sigmoid": SIGMOID.replace("\n\n", "\n"),
    "relu": RELU.replace("\n\n", "\n"),
    "identity": None,
    "avg": None,
    "max": None,
    "min": None,
}



from neuralogic.core.builder.builder import Builder
from neuralogic.core.builder.dataset_builder import DatasetBuilder
from neuralogic.core.builder.components import Neuron, Weight, Sample, BuiltDataset

__all__ = ["Builder", "DatasetBuilder", "Neuron", "Weight", "Sample", "BuiltDataset"]



from typing import Union

from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.function import Transformation, Combination, Function, Aggregation


# Transformation


def sigmoid(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SIGMOID(entity)


def tanh(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.TANH(entity)


def signum(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SIGNUM(entity)


def relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.RELU(entity)


def leaky_relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LEAKY_RELU(entity)


def lukasiewicz(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LUKASIEWICZ(entity)


def exp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.EXP(entity)


def sqrt(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SQRT(entity)


def inverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.INVERSE(entity)


def reverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.REVERSE(entity)


def log(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LOG(entity)


def identity(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.IDENTITY(entity)


def transp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.TRANSP(entity)


def softmax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SOFTMAX(entity)


def sparsemax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SPARSEMAX(entity)


def norm(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.NORM(entity)


# Combination


def max_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.MAX(entity)


def min_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.MIN(entity)


def avg_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.AVG(entity)


def sum_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.SUM(entity)


def count_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.COUNT(entity)


def product_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.PRODUCT(entity)


def elproduct_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.ELPRODUCT(entity)


def softmax_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.SOFTMAX(entity)


def sparsemax_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.SPARSEMAX(entity)


def crosssum_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.CROSSSUM(entity)


def concat_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.CONCAT(entity)


def cossim_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.COSSIM(entity)


# Aggregations


def max(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MAX(entity)


def min(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MIN(entity)


def avg(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.AVG(entity)


def sum(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.SUM(entity)


def count(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.COUNT(entity)



from typing import Union, Iterable, Callable

from neuralogic.core.constructs.function import Transformation, Combination, Aggregation, Function


class Metadata:
    __slots__ = "learnable", "transformation", "aggregation", "duplicit_grounding", "combination"

    def __init__(
        self,
        learnable: bool = None,
        transformation: Union[str, Transformation, Combination] = None,
        combination: Union[str, Combination] = None,
        aggregation: Union[str, Aggregation] = None,
        duplicit_grounding: bool = False,
    ):
        self.learnable = learnable
        self.combination = combination
        self.transformation = transformation
        self.aggregation = aggregation
        self.duplicit_grounding = duplicit_grounding

    @staticmethod
    def from_iterable(iterable: Iterable) -> "Metadata":
        metadata = Metadata()

        for entry in iterable:
            if isinstance(entry, Callable) and not isinstance(entry, Function):
                entry = entry()
            if isinstance(entry, Aggregation):
                metadata.aggregation = entry
            elif isinstance(entry, Transformation):
                metadata.transformation = entry
            elif isinstance(entry, Combination):
                metadata.combination = entry
            else:
                raise ValueError(f"Invalid entry for metadata: {entry}")
        return metadata

    def __str__(self):
        metadata_list = []
        if self.learnable is not None:
            metadata_list.append(f"learnable={str(self.learnable).lower()}")
        if self.transformation is not None:
            metadata_list.append(f"transformation={str(self.transformation)}")
        if self.combination is not None:
            metadata_list.append(f"combination={str(self.combination)}")
        if self.aggregation is not None:
            metadata_list.append(f"aggregation={str(self.aggregation)}")
        return f"[{', '.join(metadata_list)}]"



from typing import List, Union, Optional, Tuple

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core import Template, Settings, R
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.rule import Rule


class InferenceEngine:
    def __init__(self, template: Template, settings: Settings = None):
        if not is_initialized():
            initialize()

        self.settings = Settings().create_disconnected_proxy() if settings is None else settings.create_proxy()
        self.java_factory = JavaFactory()

        self.settings.settings.inferTemplateFacts = False

        self.parsed_template = template.get_parsed_template(self.settings, self.java_factory)
        self.dataset_builder = DatasetBuilder(self.parsed_template, self.java_factory)

        self.examples: List[Union[BaseRelation, Rule]] = []

        self.grounder = jpype.JClass("cz.cvut.fel.ida.logic.grounding.Grounder").getGrounder(self.settings.settings)
        self.matching = jpype.JClass("cz.cvut.fel.ida.logic.subsumption.Matching")()
        self.examples_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.ExamplesBuilder")
        self.queries_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.QueriesBuilder")
        self.grounding_sample = jpype.JClass("cz.cvut.fel.ida.logic.grounding.GroundingSample")
        self.horn_clause = jpype.JClass("cz.cvut.fel.ida.logic.HornClause")

        self.empty_example = jpype.JClass("cz.cvut.fel.ida.logic.constructs.example.LiftedExample")()

    def set_knowledge(self, examples: List[Union[BaseRelation, Rule]]) -> None:
        self.examples = examples

    def get_queries(self, examples: Optional[List[Union[BaseRelation, Rule]]] = None):
        if examples is None:
            examples = self.examples

        examples_builder = self.examples_builder(self.settings.settings)

        self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
        built_examples = self.dataset_builder.build_examples([examples], examples_builder)[0]
        sample = built_examples[0]

        gs = self.grounding_sample(sample, self.parsed_template)

        lifted_example = gs.query.evidence
        template = gs.template

        ground_template = self.grounder.groundRulesAndFacts(lifted_example, template)

        ground_rules = ground_template.groundRules.values()
        for ground_rule in ground_rules:
            for head in ground_rule.keys():
                ground_head = head.groundHead

                yield R.get(str(ground_head.predicateName()))([str(term.name()) for term in ground_head.arguments()])

    def q(self, query: BaseRelation, examples: Optional[List[Union[BaseRelation, Rule]]] = None):
        return self.query(query, examples)

    def query(self, query: BaseRelation, examples: Optional[List[Union[BaseRelation, Rule]]] = None):
        if examples is None:
            examples = self.examples

        examples_builder = self.examples_builder(self.settings.settings)
        query_builder = self.queries_builder(self.settings.settings)
        query_builder.setFactoriesFrom(examples_builder)

        self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
        queries, one_query_per_example = self.dataset_builder.build_queries([query], query_builder)

        self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
        examples = self.dataset_builder.build_examples([examples], examples_builder)[0]

        logic_samples = self.dataset_builder.merge_queries_with_examples(
            queries, examples, examples_builder, one_query_per_example
        )
        sample = logic_samples[0]

        gs = self.grounding_sample(sample, self.parsed_template)

        lifted_example = gs.query.evidence
        template = gs.template

        ground_template = self.grounder.groundRulesAndFacts(lifted_example, template)

        clause = self.java_factory.atom_to_clause(query)
        name = str(query.predicate)
        results = []
        variables = [(index, term) for index, term in enumerate(query.terms) if str(term)[0].isupper()]

        self._get_substitutions(clause, name, variables, ground_template.groundRules, results)
        self._get_substitutions(clause, name, variables, ground_template.groundFacts, results)

        if len(results) == 0:
            return {}

        if len(variables) == 0:
            return iter([])

        return results

    def _get_substitutions(
        self, clause, query_signature: str, variables: List[Tuple[int, str]], literals, substitutions: List
    ):
        for literal in literals:
            if str(literal.predicate().toString()) == query_signature and self.matching.subsumption(
                clause, self.java_factory.clause(literal)
            ):
                terms = literal.arguments()
                substitutions.append({str(label): str(terms[index]) for index, label in variables})



from typing import Optional, List

from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class MAGNNMean(Module):
    r"""
    Intra-metapath Aggregation module with Mean encoder from
    `"MAGNN: Metapath Aggregated Graph Neural Network for Heterogeneous Graph Embedding" <https://arxiv.org/abs/2002.01680>`_.
    Which can be expressed as:

    .. math::

        \mathbf{h}_{P(v,u)} = MEAN(\{\mathbf{x}_t |  \forall t \in P(v,u) \})

    .. math::

        \mathbf{h}^P_{v} = act(\sum_{u \in N^P_v} \mathbf{h}_{P(v,u)})

    Where *act* is an activation function, :math:`P(v,u)` is a single metapath instance, :math:`N^P_{v}` is set of
    metapath-based neighbors.

    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    relation_name : str
        Relation predicate name for connectivity checks between entities.
    type_name : Optional[str]
        Metapath type predicate name. If none, ``meta_paths`` will be used instead.
    meta_paths : List[str]
        Name of types forming a single metapath.
    activation : Transformation
        Activation function of the output.
        Default: ``Transformation.SIGMOID``
    """

    def __init__(
        self,
        output_name: str,
        feature_name: str,
        relation_name: str,
        type_name: Optional[str],
        meta_paths: List[str],
        activation: Transformation = Transformation.SIGMOID,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.relation_name = relation_name
        self.type_name = type_name
        self.meta_paths = meta_paths

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        metadata = Metadata(duplicit_grounding=True, transformation=Transformation.IDENTITY)
        length = len(self.meta_paths)
        feature = R.get(self.feature_name)
        relation = R.get(self.relation_name)

        meta_paths = [relation(f"V{index}", f"V{index + 1}") for index in range(len(self.meta_paths) - 1)]

        if self.type_name is None:
            meta_paths.extend(R.get(type)(f"V{index}") for index, type in enumerate(self.meta_paths))
        else:
            type_relation = R.get(self.type_name)
            meta_paths.extend(type_relation(f"V{index}", type) for index, type in enumerate(self.meta_paths))

        if length == 0:
            meta_paths.extend(feature(f"V{index}") for index in range(len(self.meta_paths)))
        else:
            meta_paths.extend(feature(f"V{index}")[1 / length].fixed() for index in range(len(self.meta_paths)))

        return [
            (R.get(self.output_name)(V.V0) <= meta_paths) | metadata,
            R.get(self.output_name) / 1 | Metadata(transformation=self.activation),
        ]


class MAGNNLinear(MAGNNMean):
    r"""
    Intra-metapath Aggregation module with Linear encoder from
    `"MAGNN: Metapath Aggregated Graph Neural Network for Heterogeneous Graph Embedding" <https://arxiv.org/abs/2002.01680>`_.
    Which can be expressed as:

    .. math::

        \mathbf{h}_{P(v,u)} = \mathbf{W}_p \cdot MEAN(\{\mathbf{x}_t |  \forall t \in P(v,u) \})

    .. math::

        \mathbf{h}^P_{v} = act(\sum_{u \in N^P_v} \mathbf{h}_{P(v,u)})

    Where *act* is an activation function, :math:`P(v,u)` is a single metapath instance, :math:`N^P_{v}` is set of
    metapath-based neighbors.

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    relation_name : str
        Relation predicate name for connectivity checks between entities.
    type_name : Optional[str]
        Metapath type predicate name. If none, ``meta_paths`` will be used instead.
    meta_paths : List[str]
        Name of types forming a single metapath.
    activation : Transformation
        Activation function of the output.
        Default: ``Transformation.SIGMOID``
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        relation_name: str,
        type_name: Optional[str],
        meta_paths: List[str],
        activation: Transformation = Transformation.SIGMOID,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        super().__init__(output_name, feature_name, relation_name, type_name, meta_paths, activation, aggregation)
        self.in_channels = in_channels
        self.out_channels = out_channels

    def __call__(self):
        rules = super().__call__()

        rules[0].head = rules[0].head[self.out_channels, self.in_channels]
        return rules



from typing import Union, List, Optional, Set, Dict, Any, Callable, Iterable

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core.builder import Builder, DatasetBuilder
from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.core.constructs.predicate import PredicateMetadata
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.settings import SettingsProxy, Settings
from neuralogic.nn.module.module import Module

from neuralogic.utils.visualize import draw_model


TemplateEntries = Union[BaseRelation, WeightedRelation, Rule]


class Template:
    def __init__(
        self,
        *,
        template_file: Optional[str] = None,
    ):
        self.template: List[TemplateEntries] = []
        self.template_file = template_file

        self.counter = 0
        self.hooks: Dict[str, Set] = {}

    def add_hook(self, relation: Union[BaseRelation, str], callback: Callable[[Any], None]) -> None:
        """Hooks the callable to be called with the relation's value as an argument when the value of
        the relation is being calculated.

        :param relation:
        :param callback:
        :return:
        """
        name = str(relation)

        if isinstance(relation, BaseRelation):
            name = name[:-1]

        if name not in self.hooks:
            self.hooks[name] = {callback}
        else:
            self.hooks[name].add(callback)

    def remove_hook(self, relation: Union[BaseRelation, str], callback):
        """Removes the callable from the relation's hooks

        :param relation:
        :param callback:
        :return:
        """
        name = str(relation)

        if isinstance(relation, BaseRelation):
            name = name[:-1]

        if name not in self.hooks:
            return
        self.hooks[name].discard(callback)

    def add_rule(self, rule) -> None:
        """Adds one rule to the template

        :param rule:
        :return:
        """
        self.add_rules([rule])

    def add_rules(self, rules: List):
        """Adds multiple rules to the template

        :param rules:
        :return:
        """
        self.template.extend(rules)

    def add_module(self, module: Module):
        """Expands the module into rules and adds them into the template

        :param module:
        :return:
        """
        self.add_rules(module())

    def get_parsed_template(self, settings: SettingsProxy, java_factory: JavaFactory):
        if not is_initialized():
            initialize()

        if self.template_file is not None:
            return Builder(settings).build_template_from_file(settings, self.template_file)

        predicate_metadata = []
        weighted_rules = []
        valued_facts = []

        for rule in self.template:
            if isinstance(rule, PredicateMetadata):
                predicate_metadata.append(java_factory.get_predicate_metadata_pair(rule))
            elif isinstance(rule, Rule):
                weighted_rules.append(java_factory.get_rule(rule))
            elif isinstance(rule, (WeightedRelation, BaseRelation)):
                valued_facts.append(java_factory.get_valued_fact(rule, java_factory.get_variable_factory()))

        parsed_template = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.types.ParsedTemplate")
        template = parsed_template(jpype.java.util.ArrayList(weighted_rules), jpype.java.util.ArrayList(valued_facts))

        template.weightsMetadata = (jpype.java.util.List) @ jpype.java.util.ArrayList([])
        template.predicatesMetadata = jpype.java.util.ArrayList(predicate_metadata)

        metadata_processor = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.transforming.MetadataProcessor")
        metadata_processor = metadata_processor(settings.settings)

        metadata_processor.processMetadata(template)
        template.inferTemplateFacts()

        return template

    def remove_duplicates(self):
        """Remove duplicates from the template"""
        entries = set()
        deduplicated_template: List[TemplateEntries] = []

        for entry in self.template:
            entry_str = str(entry)

            if entry_str in entries:
                continue
            entries.add(entry_str)
            deduplicated_template.append(entry)
        self.template = deduplicated_template

    def build(self, settings: Settings):
        from neuralogic.nn import get_neuralogic_layer

        java_factory = JavaFactory()
        settings_proxy = settings.create_proxy()

        parsed_template = self.get_parsed_template(settings_proxy, java_factory)
        model = Builder(settings_proxy).build_model(parsed_template, settings_proxy)

        return get_neuralogic_layer()(model, DatasetBuilder(parsed_template, java_factory), self, settings_proxy)

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        from neuralogic.nn import get_neuralogic_layer

        settings_proxy = Settings().create_proxy()
        java_factory = JavaFactory()

        parsed_template = self.get_parsed_template(settings_proxy, java_factory)
        model = Builder(settings_proxy).build_model(parsed_template, settings_proxy)
        layer = get_neuralogic_layer()(model, DatasetBuilder(parsed_template, java_factory), self, settings_proxy)

        return draw_model(layer, filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs)

    def __str__(self) -> str:
        return "\n".join(str(r) for r in self.template)

    def __iadd__(self, other) -> "Template":
        if isinstance(other, Iterable):
            self.template.extend(other)
        elif isinstance(other, Module):
            self.template.extend(other())
        else:
            self.template.append(other)
        return self



class Optimizer:
    @property
    def lr(self) -> float:
        raise NotImplementedError

    def get(self):
        raise NotImplementedError

    def name(self) -> str:
        return self.__class__.__name__

    def is_default(self) -> bool:
        return True



from typing import List, Union, Optional

from neuralogic.core import Template, Settings
from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.rule import Rule

from neuralogic.dataset import Dataset


class EvaluationInferenceEngine:
    def __init__(self, template: Template, settings: Settings = None):
        self.settings = Settings() if settings is None else settings
        self.model = template.build(self.settings)

        self.examples: List[Union[BaseRelation, Rule]] = []
        self.dataset = Dataset()
        self.dataset.examples = [[]]

    def set_knowledge(self, examples: List[Union[BaseRelation, Rule]]) -> None:
        self.dataset.examples = [examples]

    def q(self, query: BaseRelation, examples: Optional[List[Union[BaseRelation, Rule]]] = None):
        return self.query(query, examples)

    def query(self, query: BaseRelation, examples: Optional[List[Union[BaseRelation, Rule]]] = None):
        global_examples = self.dataset.examples

        if examples is not None:
            self.dataset.examples = [examples]

        self.dataset.queries = [query]
        variables = [(name, index) for index, name in enumerate(query.terms) if str(name).isupper()]

        try:
            built_dataset = self.model.build_dataset(self.dataset)
            results = self.model(built_dataset.samples, train=False)
        except Exception:
            self.dataset.examples = global_examples
            return {}

        self.dataset.examples = global_examples

        if len(built_dataset.samples) != len(results):
            raise Exception

        def generator():
            for result, sample in zip(results, built_dataset.samples):
                sub_query = str(sample.java_sample.query.neuron.getName())
                sub_query = sub_query.split("(")[1].strip()[:-1]

                substitutions = sub_query.split(",")
                yield result, {label: substitutions[position].strip() for label, position in variables}

        return generator()



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module


class Linear(Module):
    r"""
    Apply linear transformation on the input. Can be expressed as:

    .. math::

        h_{i_0, .., i_{n}} = W \cdot x_{i_0, .., i_{n}}

    Where :math:`x` is the input, :math:`W \in R^{(out\_channels \times in\_channels)}` is a learnable parameter,
    and :math:`n` is the arity of the input and output.

    It is also possible to attach non-linearity via the activation parameter and compute:

    .. math::

        h_{i_0, .., i_{n}} = act(W \cdot x_{i_0, .., i_{n}})

    Example
    -------

    The whole computation of this module (parametrized as :code:`Linear(1, 2, "h1", "h0")`) is as follows:

    .. code:: logtalk

        (R.h1(V.X0)[2, 1] <= R.h0(V.X0)) | [Transformation.IDENTITY]
        R.h1 / 1 | [Transformation.IDENTITY]

    Module parametrized as :code:`Linear(1, 2, "h1", "h0", Transformation.SIGMOID, 2)` translates into:

    .. code:: logtalk

        (R.h1(V.X0, V.X1)[2, 1] <= R.h0(V.X0, V.X1)) | [Transformation.IDENTITY]
        R.h1 / 2 | [Transformation.SIGMOID]

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    activation : Transformation
        Activation function of the output.
        Default: ``Transformation.IDENTITY``
    arity : int
        Arity of the input and output predicate. Default: ``1``
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        input_name: str,
        activation: Transformation = Transformation.IDENTITY,
        arity: int = 1,
    ):
        self.output_name = output_name
        self.input_name = input_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.activation = activation
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        head = R.get(self.output_name)(terms)[self.out_channels, self.in_channels]

        return [
            (head <= R.get(self.input_name)(terms)) | [Transformation.IDENTITY],
            R.get(self.output_name) / len(terms) | Metadata(transformation=self.activation),
        ]



from typing import Optional, Iterable, Sequence

import numpy as np
import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core.settings import SettingsProxy, Settings


class ValueFactory:
    def __init__(self):
        self.scalar_value = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")
        self.vector_value = jpype.JClass("cz.cvut.fel.ida.algebra.values.VectorValue")
        self.matrix_value = jpype.JClass("cz.cvut.fel.ida.algebra.values.MatrixValue")

    def get_value(self, weight):
        if np.ndim(weight) == 0:
            return True, self.scalar_value(float(weight))

        if isinstance(weight, tuple):
            if len(weight) == 1:
                if weight[0] == 1:
                    value = self.scalar_value()
                else:
                    value = self.vector_value(weight[0])
            elif len(weight) == 2:
                if weight[0] == 1:
                    value = self.vector_value(weight[1], True)
                elif weight[1] == 1:
                    value = self.vector_value(weight[0], False)
                else:
                    value = self.matrix_value(weight[0], weight[1])
            else:
                raise NotImplementedError
            return False, value

        if isinstance(weight, (Sequence, np.ndarray, Iterable)):
            if len(weight) == 0:
                raise NotImplementedError

            if np.ndim(weight[0]) == 0:
                vector = [float(w) for w in weight]
                return True, self.vector_value(vector)

            if isinstance(weight[0], (Sequence, np.ndarray, Iterable)):
                if len(weight) == 1:
                    value = self.vector_value([float(w) for w in weight[0]])
                    value.rowOrientation = True
                else:
                    try:
                        matrix = [float(w) for weights in weight for w in weights]
                        value = self.matrix_value(matrix, len(weight), len(weight[0]))
                    except TypeError:
                        value = self.vector_value([float(w) for w in weight])
                return True, value

        raise ValueError(f"Cannot create weight from type {type(weight)}, value {weight}")


class JavaFactory:
    def __init__(self, settings: Optional[SettingsProxy] = None):
        from neuralogic.core.constructs.rule import Rule
        from neuralogic.core.constructs.relation import WeightedRelation

        if not is_initialized():
            initialize()

        if settings is None:
            settings = Settings().create_proxy()

        self.settings = settings

        self.value_factory = ValueFactory()

        self.weighted_atom_type = WeightedRelation
        self.rule_type = Rule

        self.predicate_metadata = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.metadata.PredicateMetadata")
        self.rule_metadata = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.metadata.RuleMetadata")

        self.template_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.TemplateBuilder")
        self.examples_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.ExamplesBuilder")

        self.var_factory_class = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.factories.VariableFactory")

        self.body_atom = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.components.BodyAtom")
        self.head_atom = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.components.HeadAtom")
        self.weighted_rule = jpype.JClass("cz.cvut.fel.ida.logic.constructs.template.components.WeightedRule")

        self.builder = self.template_builder(settings.settings)

        self.constant_factory = self.builder.constantFactory
        self.predicate_factory = self.builder.predicateFactory
        self.weight_factory = self.builder.weightFactory

        self.lifted_example = jpype.JClass("cz.cvut.fel.ida.logic.constructs.example.LiftedExample")
        self.valued_fact = jpype.JClass("cz.cvut.fel.ida.logic.constructs.example.ValuedFact")

        self.clause = jpype.JClass("cz.cvut.fel.ida.logic.Clause")
        self.literal = jpype.JClass("cz.cvut.fel.ida.logic.Literal")
        self.conjunction = jpype.JClass("cz.cvut.fel.ida.logic.constructs.Conjunction")

        self.string_value = jpype.JClass("cz.cvut.fel.ida.algebra.values.StringValue")

        self.transformation = jpype.JClass("cz.cvut.fel.ida.algebra.functions.Transformation")

        self.pair = jpype.JClass("cz.cvut.fel.ida.utils.generic.Pair")
        self.settings_class = settings.settings_class

        self.unit_weight = jpype.JClass("cz.cvut.fel.ida.algebra.weights.Weight").unitWeight
        self.variable_factory = self.get_variable_factory()

    def get_variable_factory(self):
        return self.var_factory_class()

    def get_term(self, term, variable_factory):
        if isinstance(term, str):
            if term[0].islower() or term.isnumeric():
                return self.constant_factory.construct(term)
            elif term[0].isupper():
                return variable_factory.construct(term)
            else:
                raise ValueError(f"Invalid term {term}")
        if isinstance(term, (int, float)):
            return self.constant_factory.construct(str(term))
        raise ValueError(f"Invalid term {term}")

    def atom_to_clause(self, atom):
        terms = [self.get_term(term, self.variable_factory) for term in atom.terms]

        predicate_name = f"@{atom.predicate.name}" if atom.predicate.special else atom.predicate.name
        literal = self.literal(predicate_name, False, terms)
        literal_array = self.literal[1]
        literal_array[0] = literal

        return self.clause(literal_array)

    def get_generic_relation(self, relation_class, relation, variable_factory, default_weight=None, is_example=False):
        predicate = self.get_predicate(relation.predicate)

        weight = None
        if isinstance(relation, self.weighted_atom_type):
            weight = self.get_weight(relation.weight, relation.weight_name, relation.is_fixed or is_example)
        elif default_weight is not None:
            weight = self.get_weight(default_weight, None, True)

        term_list = [self.get_term(term, variable_factory) for term in relation.terms]

        j_term_list = jpype.java.util.ArrayList()
        for x in term_list:
            j_term_list.add(x)

        if relation_class == self.body_atom:
            function = self.settings_class.parseTransformation(str(relation.function)) if relation.function else None
            transformation_function = self.transformation.getFunction(function) if function is not None else None

            java_relation = relation_class(predicate, j_term_list, False, transformation_function, weight)
        else:
            java_relation = relation_class(predicate, j_term_list, False, weight)
        java_relation.originalString = relation.to_str()

        return java_relation

    def get_metadata(self, metadata, metadata_class):
        if metadata is None:
            return None

        if (
            metadata.aggregation is None
            and metadata.transformation is None
            and metadata.combination is None
            and metadata.learnable is None
        ):
            return None

        map = jpype.JClass("java.util.LinkedHashMap")()

        if metadata.aggregation is not None:
            map.put("aggregation", self.string_value(str(metadata.aggregation).lower()))
        if metadata.transformation is not None:
            map.put("transformation", self.string_value(str(metadata.transformation).lower()))
        if metadata.combination is not None:
            map.put("combination", self.string_value(str(metadata.combination).lower()))
        if metadata.learnable is not None:
            map.put("learnable", self.string_value(str(metadata.learnable).lower()))

        return metadata_class(self.builder.settings, map)

    def get_query(self, query):
        variable_factory = self.get_variable_factory()

        if not isinstance(query, self.rule_type):
            if not isinstance(query, Iterable):
                query = [query]
            return None, self.get_conjunction(query, variable_factory, 1.0, True)
        return self.get_relation(query.head, variable_factory, True), self.get_conjunction(
            query.body, variable_factory, is_example=True
        )

    def get_lifted_example(self, example, learnable_facts=False):
        conjunctions = []
        rules = []
        label_conjunction = None

        variable_factory = self.get_variable_factory()

        if not isinstance(example, self.rule_type):
            if not isinstance(example, Iterable):
                example = [example]
            conjunctions.append(self.get_conjunction(example, variable_factory, is_example=not learnable_facts))
        else:
            label_conjunction = self.get_conjunction([example.head], variable_factory, is_example=not learnable_facts)
            conjunctions.append(self.get_conjunction(example.body, variable_factory, is_example=not learnable_facts))

        lifted_example = self.lifted_example(jpype.java.util.ArrayList(conjunctions), jpype.java.util.ArrayList(rules))
        return label_conjunction, lifted_example

    def get_conjunction(self, relations, variable_factory, default_weight=None, is_example=False):
        valued_facts = [
            self.get_valued_fact(relation, variable_factory, default_weight, is_example) for relation in relations
        ]
        return self.conjunction(jpype.java.util.ArrayList(valued_facts))

    def get_predicate_metadata_pair(self, predicate_metadata):
        return self.pair(
            self.get_predicate(predicate_metadata.predicate),
            self.get_metadata(predicate_metadata.metadata, self.predicate_metadata),
        )

    def get_valued_fact(self, relation, variable_factory, default_weight=None, is_example=False):
        return self.get_generic_relation(
            self.valued_fact,
            relation,
            variable_factory,
            default_weight,
            is_example,
        )

    def get_relation(self, relation, variable_factory, is_example=False):
        return self.get_generic_relation(self.body_atom, relation, variable_factory, is_example=is_example)

    def get_rule(self, rule):
        java_rule = self.weighted_rule()
        java_rule.setOriginalString(str(rule))

        variable_factory = self.get_variable_factory()

        head_relation = self.get_relation(rule.head, variable_factory)
        weight = head_relation.getConjunctWeight()

        if weight is None:
            java_rule.setWeight(self.unit_weight)
        else:
            java_rule.setWeight(weight)

        body_relation = [self.get_relation(relation, variable_factory) for relation in rule.body]
        body_relation_list = jpype.java.util.ArrayList(body_relation)

        java_rule.setHead(self.head_atom(head_relation))
        java_rule.setBody(body_relation_list)

        offset = None  # TODO: Implement

        java_rule.setOffset(offset)

        if rule.metadata is not None:
            java_rule.allowDuplicitGroundings = rule.metadata.duplicit_grounding

        java_rule.setMetadata(self.get_metadata(rule.metadata, self.rule_metadata))

        return java_rule

    def get_predicate(self, predicate):
        return self.predicate_factory.construct(predicate.name, predicate.arity, predicate.special, predicate.hidden)

    def get_weight(self, weight, name, fixed):
        initialized, value = self.value_factory.get_value(weight)

        if name is None:
            return self.weight_factory.construct(value, fixed, initialized)
        return self.weight_factory.construct(name, value, fixed, initialized)

    def get_new_weight_factory(self):
        return self.examples_builder(self.settings.settings).weightFactory



import jpype

from neuralogic.optim.optimizer import Optimizer


class SGD(Optimizer):
    def __init__(self, lr: float = 0.1):
        self._lr = lr

    @property
    def lr(self) -> float:
        return self._lr

    def get(self):
        sgd_class = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.optimizers.SGD")
        lr = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")(self._lr)

        return sgd_class(lr)

    def __str__(self) -> str:
        return f"SGD(lr={self.lr})"



from typing import List, Optional

import jpype
from halo._utils import get_environment

from neuralogic import is_initialized, initialize
from neuralogic.core.builder.components import Sample, RawSample
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.sources import Sources


def stream_to_list(stream) -> List:
    return list(stream.collect(jpype.JClass("java.util.stream.Collectors").toList()))


class Builder:
    def __init__(self, settings: SettingsProxy):
        if not is_initialized():
            initialize()

        self.settings = settings
        self.example_builder = Builder.get_builders(settings)
        self.builder = Builder.get_builders(settings)

        self.neural_model = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.NeuralModel")
        self.collectors = jpype.JClass("java.util.stream.Collectors")

        @jpype.JImplements(jpype.JClass("java.util.function.IntConsumer"))
        class Callback:
            def __init__(self, spinner):
                self.state = 0
                self.spinner = spinner

            @jpype.JOverride
            def accept(self, count: int):
                self.state = count
                self.spinner.text = Builder._get_spinner_text(count)

        self._callback = Callback

    def build_template_from_file(self, settings: SettingsProxy, filename: str):
        args = [
            "-t",
            filename,
            "-q",
            filename,
        ]

        sources = Sources.from_args(args, settings)
        template = self.builder.buildTemplate(sources.sources)

        return template

    def from_sources(self, parsed_template, sources: Sources, progress: bool) -> List[RawSample]:
        if progress:
            return self._build_samples_with_progress(parsed_template, sources, None)
        return self._build_samples(parsed_template, sources, None)

    def from_logic_samples(self, parsed_template, logic_samples, progress: bool) -> List[RawSample]:
        if progress:
            return self._build_samples_with_progress(parsed_template, None, logic_samples)
        return self._build_samples(parsed_template, None, logic_samples)

    def _build_samples_with_progress(
        self, parsed_template, sources: Optional[Sources], logic_samples
    ) -> List[RawSample]:
        jupyter = get_environment() in ("ipython", "jupyter")

        if jupyter:
            from halo import HaloNotebook as Halo
        else:
            from halo import Halo

        self.spinner = Halo(text=self._get_spinner_text(0), spinner="dots")
        self.spinner.start()

        try:
            results = self._build_samples(parsed_template, sources, logic_samples, self._callback(self.spinner))

            self.spinner.succeed()

            if jupyter:
                self.spinner.start().succeed()
            return results
        except Exception as e:
            self.spinner.fail("Building failed")

            if jupyter:
                self.spinner.start().fail()
            raise e

    def _build_samples(
        self, parsed_template, sources: Optional[Sources], logic_samples, callback=None
    ) -> List[RawSample]:
        if logic_samples is None:
            source_pipeline = self.example_builder.buildPipeline(parsed_template, sources.sources, callback)
        else:
            source_pipeline = self.example_builder.buildPipeline(parsed_template, logic_samples, callback)

        source_pipeline.execute(None if sources is None else sources.sources)
        java_model = source_pipeline.get()

        logic_samples = java_model.s
        logic_samples = logic_samples.collect(self.collectors.toList())

        return [RawSample(sample) for sample in logic_samples]

    def build_model(self, parsed_template, settings: SettingsProxy):
        neural_model = self.neural_model(parsed_template.getAllWeights(), settings.settings)

        return neural_model

    @staticmethod
    def get_builders(settings: SettingsProxy):
        builder = jpype.JClass("cz.cvut.fel.ida.pipelines.building.PythonBuilder")(settings.settings)

        return builder

    @staticmethod
    def build(samples):
        serializer = jpype.JClass("cz.cvut.fel.ida.neural.networks.structure.export.NeuralSerializer")()
        super_detailed_format = jpype.JClass("cz.cvut.fel.ida.setup.Settings").superDetailedNumberFormat
        serializer.numberFormat = super_detailed_format

        return [Sample(serializer.serialize(sample), sample) for sample in samples]

    @staticmethod
    def _get_spinner_text(count: int) -> str:
        if count == 1:
            return f"Built {count} sample"
        return f"Built {count} samples"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os

from typing import Dict
from setuptools import find_packages, setup

# Package meta-data.
NAME = "neuralogic"
DESCRIPTION = "PyNeuraLogic lets you use Python to create Differentiable Logic Programs."
URL = "https://github.com/LukasZahradnik/PyNeuraLogic"
EMAIL = "lukaszahradnik96@seznam.cz"
AUTHOR = "Lukáš Zahradník"
REQUIRES_PYTHON = ">=3.7.0"
VERSION = None

REQUIRED = [
    "JPype1 >=1.3.0",
    "numpy >= 1.19.0",
    "halo >= 0.0.31",
]

EXTRAS: Dict = {}

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about: Dict = {}

try:
    with open(os.path.join(here, NAME, "__version__.py")) as f:  # type: ignore
        exec(f.read(), about)
except Exception:
    about["__version__"] = "0.0.0"  # development

setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*", "examples"]),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    package_data={
        "neuralogic.jar": ["NeuraLogic.jar"],
        "neuralogic.utils.data": ["datasets/**/*", "datasets/**/**/*", "datasets/**/**/**/**/*", "datasets/**/**/**/*"],
    },
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



from typing import Optional, List, Union

from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.dataset.base import BaseDataset

DatasetEntries = Union[BaseRelation, WeightedRelation, Rule]


class Dataset(BaseDataset):
    r"""
    Dataset encapsulating (learning) samples in the form of logic format, allowing users to fully take advantage of the
    PyNeuraLogic library.

    One learning sample consists of:

    * Example: A list of logic facts and rules representing some instance (e.g., a graph)

    * Query: A logic fact to mark the output of a model and optionally target label.

    Examples and queries in the dataset can be paired in the following ways:

    * :math:`N:N` - Dataset contains :math:`N` examples and :math:`N` queries. They will be paired by their index.

    .. code:: python

        dataset.add_example(first_example)
        dataset.add_example(second_example)

        dataset.add_query(first_query)
        dataset.add_query(second_query)

        # Learning samples: [first_example, first_query], [second_example, second_query]

    * :math:`1:N` - Dataset contains :math:`1` example and :math:`N` queries. All queries will be run on the example.

    .. code:: python

        dataset.add_example(example)

        dataset.add_query(first_query)
        dataset.add_query(second_query)

        # Learning samples: [example, first_query], [example, second_query]

    * :math:`N:M` - Dataset contains :math:`N` examples and :math:`M` queries (:math:`N \leq M`).
      It pairs queries similarly to the :math:`N:N` case but also allows running multiple queries on a specific example
      (by inserting a list of queries instead of one query).

    .. code:: python

        dataset.add_example(first_example)
        dataset.add_example(second_example)

        dataset.add_query([first_query_0, first_query_1])
        dataset.add_query(second_query)

        # Learning samples:
        #   [first_example, first_query_0], [first_example, first_query_1], [second_example, second_query]

    Parameters
    ----------

    examples : Optional[List]
        List of examples. Default: ``None``
    queries : Optional[List]
        List of queries. Default: ``None``

    """

    def __init__(
        self,
        examples: Optional[List[List[DatasetEntries]]] = None,
        queries: Optional[List[Union[List[DatasetEntries], DatasetEntries]]] = None,
    ):
        self.examples: List[List[DatasetEntries]] = examples if examples is not None else []
        self.queries: List[Union[List[DatasetEntries], DatasetEntries]] = queries if queries is not None else []

    def add_example(self, example):
        self.add_examples([example])

    def add_examples(self, examples: List):
        self.examples.extend(examples)

    def set_examples(self, examples: List):
        self.examples = examples

    def add_query(self, query):
        self.add_queries([query])

    def add_queries(self, queries: List):
        self.queries.extend(queries)

    def set_queries(self, queries: List):
        self.queries = queries

    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        for examples in self.examples:
            examples_fp.write(f"{','.join(example.to_str(False) for example in examples)}.{sep}")

        for query in self.queries:
            queries_fp.write(f"{query}{sep}")
